#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re
import json
import paramiko
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState




from .abc_test import AbcTest

from .. import images

class LinkBenchmark(AbcTest):

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

    @staticmethod
    def latency_test(clientn1, clientn2, ip_of_interface_on_n1, ip_of_interface_on_n2, verbose):
        output = "Information about latency with ping: \n"
        stdin, stdout, stderr = clientn1.exec_command('ping -c 5 ' + ip_of_interface_on_n2 + ' | grep rtt')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn2.exec_command('ping -c 5 ' + ip_of_interface_on_n1 + ' | grep rtt')
        output += "\n" + stdout.read().decode("utf-8")

        return {"Latency" : output}

    @staticmethod
    def mtu_test(clientn1, clientn2, ip_of_interface_on_n1, ip_of_interface_on_n2, verbose):
        output = "Information about mtu with ping: \n"
        ping_packets_count = 3
        ping_packet_sizes = [9000, 8950, 8000, 1500, 1450, 1400, 1000, 500, 100, 50]
        for ping_packet_size in ping_packet_sizes:
            stdin, stdout, stderr = clientn1.exec_command('ping -M do -s ' + str(ping_packet_size) + ' -c ' + str(ping_packets_count) + ' ' + ip_of_interface_on_n2)
            ping_string = stdout.read().decode("utf-8")
            ping_string = re.findall("[0-9] received", ping_string)
            ping_string = re.findall("[0-9]", ping_string[0])
            if(int(ping_string[0]) == ping_packets_count):
                output += "Packet size " + str(ping_packet_size + 8) + " is enabled."
                break
            else:
                if(verbose):
                    print("Packet " + str(ping_packet_size + 8) + " too large.")
        for ping_packet_size in ping_packet_sizes:
            stdin, stdout, stderr = clientn2.exec_command('ping -M do -s ' + str(ping_packet_size) + ' -c ' + str(ping_packets_count) + ' ' + ip_of_interface_on_n1)
            ping_string = stdout.read().decode("utf-8")
            ping_string = re.findall("[0-9] received", ping_string)
            ping_string = re.findall("[0-9]", ping_string[0])
            if(int(ping_string[0]) == ping_packets_count):
                output += "\n" + "Packet size " + str(ping_packet_size + 8) + " is enabled."
                break
            else:
                if(verbose):
                    print("Packet " + str(ping_packet_size + 8) + " too large.")

        return {"MTU" : output}

    @staticmethod
    def bandwidth_test(clientn1, clientn2, ip_of_interface_on_n1, ip_of_interface_on_n2, verbose):
        output = "Information about bandwidth with iperf: \n"
        stdin, stdout, stderr = clientn1.exec_command('echo "net.core.rmem_max = 2147483647\nnet.core.wmem_max = 2147483647\nnet.ipv4.tcp_rmem = 4096 87380 2147483647\nnet.ipv4.tcp_wmem = 4096 65536 2147483647\nnet.ipv4.tcp_congestion_control=htcp\nnet.ipv4.tcp_mtu_probing=1\nnet.core.default_qdisc = fq\n" | sudo tee -a /etc/sysctl.conf && sudo sysctl -p')
        stdin, stdout, stderr = clientn2.exec_command('echo "net.core.rmem_max = 2147483647\nnet.core.wmem_max = 2147483647\nnet.ipv4.tcp_rmem = 4096 87380 2147483647\nnet.ipv4.tcp_wmem = 4096 65536 2147483647\nnet.ipv4.tcp_congestion_control=htcp\nnet.ipv4.tcp_mtu_probing=1\nnet.core.default_qdisc = fq\n" | sudo tee -a /etc/sysctl.conf && sudo sysctl -p')
        stdin, stdout, stderr = clientn1.exec_command('iperf3 -s > /dev/null 2>&1 &')
        stdin, stdout, stderr = clientn2.exec_command('iperf3 -s > /dev/null 2>&1 &')
        stdin, stdout, stderr = clientn2.exec_command('iperf3 -c ' + ip_of_interface_on_n1 + ' -P 32 -w 512M -R')
        iperf_string = stdout.read().decode("utf-8")
        output += "n2 to n1:\n"
        iperf_strings = iperf_string.splitlines()
        if(len(iperf_strings) > 3):
            output += iperf_string.splitlines()[-4] + "\n"
            output += iperf_string.splitlines()[-3] + "\n"
        stdin, stdout, stderr = clientn1.exec_command('iperf3 -c ' + ip_of_interface_on_n2 + ' -P 32 -w 512M -R')
        iperf_string = stdout.read().decode("utf-8")
        # iperf_string2 = re.findall("^(.*)\n^(.*)\n^(.*)\n^(.*)$\z", iperf_string)
        output += "n1 to n2:\n"
        iperf_strings = iperf_string.splitlines()
        if(len(iperf_strings) > 3):
            output += iperf_string.splitlines()[-4] + "\n"
            output += iperf_string.splitlines()[-3]

        return {"Bandwidth" : output}

    @staticmethod
    def network_card_information(clientn1, clientn2, ip_of_interface_on_n1, ip_of_interface_on_n2, verbose):
        output = ""
        output += "lspci -xxxvvv\n\n"
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep PN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep V2')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep SN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep PN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep V3')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep VA')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = clientn1.exec_command('sudo lspci -xxxvvv | grep V0')
        output += stdout.read().decode("utf-8")

        output += "\n\n\nip a\n\n"
        stdin, stdout, stderr = clientn1.exec_command('ip a')
        output += stdout.read().decode("utf-8")

        return {"NetworkCardInformation" : output}

    @staticmethod
    def processor_information(clientn1, clientn2, ip_of_interface_on_n1, ip_of_interface_on_n2, verbose):
        output = ""

        output += 'sudo dmidecode | grep -w ID | sed "s/^.ID\: //g"'
        stdin, stdout, stderr = clientn1.exec_command('sudo dmidecode | grep -w ID | sed "s/^.ID\: //g"')
        output += stdout.read().decode("utf-8")

        output += '\n\n\ndmesg | grep -i dmi: | cut -d ":" -f 2-\n\n'
        stdin, stdout, stderr = clientn1.exec_command('dmesg | grep -i dmi: | cut -d ":" -f 2-')
        output += stdout.read().decode("utf-8")

        output += "\n\n\nsudo dmidecode -s system-serial-number\n\n"
        stdin, stdout, stderr = clientn1.exec_command('sudo dmidecode -s system-serial-number')
        output += stdout.read().decode("utf-8")

        return {"ProcessorInformation" : output}


    @staticmethod
    def test_ptp_accross_two_sites(site1, site2, test_list, verbose = True):
        credmgr_host = os.environ['FABRIC_CREDMGR_HOST']
        orchestrator_host = os.environ['FABRIC_ORCHESTRATOR_HOST']
        slice_manager = SliceManager(oc_host=orchestrator_host, cm_host=credmgr_host, project_name='all', scope='all')
        slice_manager.initialize()

        t = ExperimentTopology()
        cap = Capacities(core=32, ram=128, disk=10)
        n1 = t.add_node(name='n1', site=site1)
        n1.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
        n2 = t.add_node(name='n2', site=site2)
        n2.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
        n1.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name='n1-nic1')
        n2.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name='n2-nic1')
        t.add_network_service(name='ptp1', nstype=ServiceType.L2PTP, interfaces=[n1.interface_list[0], n2.interface_list[0]])
        if_labels = n1.interface_list[0].get_property(pname="labels")
        if_labels.vlan = "200"
        n1.interface_list[0].set_properties(labels=if_labels)
        if_labels = n2.interface_list[0].get_property(pname="labels")
        if_labels.vlan = "200"
        n2.interface_list[0].set_properties(labels=if_labels)
        slice_graph = t.serialize()
        ssh_key = None
        with open ("/Users/pruth/.ssh/id_rsa.pub", "r") as myfile:
            ssh_key=myfile.read()
            ssh_key=ssh_key.strip()
        status, reservations = slice_manager.create(slice_name="test_harness_latency_mtu_bandwidth", slice_graph=slice_graph, ssh_key=ssh_key)
        if(status != Status.OK):
            print(status)
            print(reservations)
            raise Exception("Slice creation failed. One thing to do: try renaming it?")
        slice_id=reservations[0].slice_id

        seconds_to_sleep = 15.0
        while True:
            return_status, slices = slice_manager.slices(excludes=[SliceState.Dead])
            if(list(filter(lambda x : x.slice_id == slice_id, slices))[0].slice_state == "StableOK"):
                if(verbose):
                    print("Slice is StableOK.")
                break
            elif(list(filter(lambda x : x.slice_id == slice_id, slices))[0].slice_state != "Configuring"):
                print("Slice state: " + list(filter(lambda x : x.slice_id == slice_id, slices))[0].slice_state)
        #         print(slice_manager.slivers(slice_id=slice_id))
                slice_object = list(filter(lambda x: x.slice_id == slice_id, slices))[0]
                status, slivers = slice_manager.slivers(slice_object=slice_object)
                for s in slivers:
                    status, sliver_status = slice_manager.sliver_status(sliver=s)
                    print("Response Status {}".format(status))
                    print("Sliver Status {}".format(sliver_status))
                raise Exception("Slice creation failed.")
            else:
                if(verbose):
                    print("Slice is configuring. Trying again in " + str(seconds_to_sleep) + " seconds.")
            time.sleep(seconds_to_sleep)

        slice_object = list(filter(lambda x: x.slice_id == slice_id, slices))[0]
        status, slivers = slice_manager.slivers(slice_object=slice_object)
        n1_ip = list(filter(lambda sliver : sliver.name == "n1",slivers))[0].management_ip
        n2_ip = list(filter(lambda sliver : sliver.name == "n2",slivers))[0].management_ip
        if(verbose):
            print("n1 IP: " + n1_ip)
            print("n2 IP: " + n2_ip)

        slice_information = "SliceID: " + slice_id + "\nSlivers:\n" + str(slivers)

        output = {"SliceInformation" : slice_information}

        key = paramiko.RSAKey.from_private_key_file("/Users/pruth/.ssh/id_rsa")
        clientn1 = paramiko.SSHClient()
        clientn1.load_system_host_keys()
        clientn1.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
        clientn1.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        connection_attempts = 0
        while True:
            try:
                clientn1.connect(n1_ip,username='ubuntu',pkey = key)
                break
            except Exception as exception:
                print(exception)
                connection_attempts += 1
                if(connection_attempts < 10):
                    if(verbose):
                        print("Connection failed. Will try to connect again in a few seconds.")
                    time.sleep(5)
                else:
                    raise Exception("Connection to server failed.")
        clientn2 = paramiko.SSHClient()
        clientn2.load_system_host_keys()
        clientn2.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
        clientn2.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        connection_attempts = 0
        while True:
            try:
                clientn2.connect(n2_ip,username='ubuntu',pkey = key)
                break
            except Exception as exception:
                print(exception)
                connection_attempts += 1
                if(connection_attempts < 10):
                    if(verbose):
                        print("Connection failed. Will try to connect again in a few seconds.")
                    time.sleep(5)
                else:
                    raise Exception("Connection to server failed.")
        def get_interface_before_last(stdout):
            interface = re.findall(r"[0-9]: [A-Za-z][A-Za-z][A-Za-z][0-9]", stdout.read().decode("utf-8"))
            interface = interface[-2:-1]
            interface = re.findall("[A-Za-z][A-Za-z][A-Za-z][0-9]", interface[0])
            return interface[0]

        stdin, stdout, stderr = clientn1.exec_command('sudo apt-get update && sudo apt-get install -y iperf iperf3')
        stdin, stdout, stderr = clientn2.exec_command('sudo apt-get update && sudo apt-get install -y iperf iperf3')

        ################################Setting up the IP addresses and activating the interfaces
        stdin, stdout, stderr = clientn1.exec_command('ip a')
        interface_n1 = get_interface_before_last(stdout)
        ip_of_interface_on_n1 = "192.168.10.51"
        stdin, stdout, stderr = clientn1.exec_command('sudo ip link add link ' + interface_n1 + ' name ens7.200 type vlan id 200')
        stdin, stdout, stderr = clientn1.exec_command('sudo ip link set dev ' + interface_n1 + ' up mtu 9000')
        stdin, stdout, stderr = clientn1.exec_command('sudo ip link set dev ens7.200 up mtu 9000')
        stdin, stdout, stderr = clientn1.exec_command('sudo ip addr add ' + ip_of_interface_on_n1 + '/24 dev ens7.200')
        stdin, stdout, stderr = clientn2.exec_command('ip a')
        interface_n2 = get_interface_before_last(stdout)
        ip_of_interface_on_n2 = "192.168.10.52"
        stdin, stdout, stderr = clientn2.exec_command('sudo ip link add link ' + interface_n2 + ' name ens7.200 type vlan id 200')
        stdin, stdout, stderr = clientn2.exec_command('sudo ip link set dev ' + interface_n2 + ' up mtu 9000')
        stdin, stdout, stderr = clientn2.exec_command('sudo ip link set dev ens7.200 up mtu 9000')
        stdin, stdout, stderr = clientn2.exec_command('sudo ip addr add ' + ip_of_interface_on_n2 + '/24 dev ens7.200')

        for test in test_list:
            output.update(test(clientn1, clientn2, ip_of_interface_on_n1, ip_of_interface_on_n2, verbose))

        slice_manager.delete(slice_object=slice_object)

        if(verbose):
            for k in output:
                print(k)
                print(output[k])
                print("---")

        return output

    def create_slice(self):
        slice_name="NVMEBenchmark"
        site_name="STAR"
        node_name='Node1'
        image = 'default_centos_8'
        image_type = 'qcow2'
        cores = 4
        ram = 8
        disk = 50
        model_type = ComponentModelType.NVME_P4510
        #nvme_component_type = ComponentType.NVME
        #nvme_model = 'P4510'
        nvme_name = 'nvme1'

        self.topology = ExperimentTopology()


        # Add node
        n1 = self.topology.add_node(name=node_name, site=site_name)

        # Set capacities
        cap = Capacities(core=cores, ram=ram, disk=disk)

        # Set Properties
        n1.set_properties(capacities=cap, image_type=image_type, image_ref=image)

        # Add the PCI NVMe device
        n1.add_component(model_type=model_type, name=nvme_name)
        #n1.add_component(ctype=nvme_component_type, model=nvme_model, name=nvme_name)

        # Generate Slice Graph
        slice_graph = self.topology.serialize()

        # Request slice from Orchestrator
        return_status, slice_reservations = self.slice_manager.create(slice_name=slice_name,
                                                    slice_graph=slice_graph,
                                                    ssh_key=self.ssh_key)




        if return_status == Status.OK:
            slice_id = slice_reservations[0].get_slice_id()
            print("Submitted slice creation request. Slice ID: {}".format(slice_id))
        else:
            print(f"Failure: {slice_reservations}")

        time.sleep(30)

        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])

        if return_status == Status.OK:
            self.slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
            self.slice = self.wait_for_slice(slice, progress=True)

        print()
        print("Slice Name : {}".format(self.slice.slice_name))
        print("ID         : {}".format(self.slice.slice_id))
        print("State      : {}".format(self.slice.slice_state))
        print("Lease End  : {}".format(self.slice.lease_end))

    def config_test(self):
        pass

    def run_test(self):
        pass

    def delete_slice(self):
        status, result = self.slice_manager.delete(slice_id=self.slice.slice_id)

        print("Response Status {}".format(status))
        print("Response received {}".format(result))


    def run(self):
        """
        Run the test
        :return:
        """
        #print( list(ComponentModelType))
        #print(self.advertised_topology)
        #self.create_slice()

        d1 = LinkBenchmark.test_ptp_accross_two_sites('TACC', 'MAX', test_list=[LinkBenchmark.latency_test,
                                                                                LinkBenchmark.mtu_test,
                                                                                LinkBenchmark.bandwidth_test,
                                                                                LinkBenchmark.network_card_information,
                                                                                LinkBenchmark.processor_information])



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, CapacityHints, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

from .abc_fablib import AbcFabLIB

from .. import images

class fablib(AbcFabLIB):
    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

        self.build_slice_manager()

    def build_slice_manager(self):
        self.slice_manager = SliceManager(oc_host=self.orchestrator_host,
                             cm_host=self.credmgr_host,
                             project_name='all',
                             scope='all')

        # Initialize the slice manager
        self.slice_manager.initialize()

        return self.slice_manager

    @staticmethod
    def init_fablib():
        if not hasattr(fablib, 'fablib_object'):
            fablib.fablib_object = fablib()

    @staticmethod
    def get_default_slice_key():
        return fablib.fablib_object.default_slice_key

    @staticmethod
    def get_config():
        return { 'credmgr_host': fablib.fablib_object.credmgr_host,
                'orchestrator_host': fablib.fablib_object.orchestrator_host,
                'fabric_token': fablib.fablib_object.fabric_token,
                'bastion_username': fablib.fablib_object.bastion_username,
                'bastion_key_filename': fablib.fablib_object.bastion_key_filename,
                'bastion_public_addr': fablib.fablib_object.bastion_public_addr,
                'bastion_public_addr': fablib.fablib_object.bastion_public_addr,
                'bastion_private_ipv4_addr': fablib.fablib_object.bastion_private_ipv4_addr,
                'slice_public_key': fablib.get_default_slice_public_key(),
                'slice_public_key_file': fablib.get_default_slice_public_key_file(),
                'slice_private_key_file': fablib.get_default_slice_private_key_file(),
                'fabric_slice_private_key_passphrase': fablib.get_default_slice_private_key_passphrase()
               }
    @staticmethod
    def get_default_slice_public_key():
        if 'slice_public_key' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_public_key']
        else:
            return None

    @staticmethod
    def get_default_slice_public_key_file():
        if 'slice_public_key_file' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_public_key_file']
        else:
            return None

    @staticmethod
    def get_default_slice_private_key_file():
        if 'slice_private_key_file' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_private_key_file']
        else:
            return None

    @staticmethod
    def get_default_slice_private_key_passphrase():
        if 'slice_private_key_passphrase' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_private_key_passphrase']
        else:
            return None

    @staticmethod
    def get_credmgr_host():
        return fablib.fablib_object.credmgr_host

    @staticmethod
    def get_orchestrator_host():
        return fablib.fablib_object.orchestrator_host

    @staticmethod
    def get_fabric_token():
        return fablib.fablib_object.fabric_token

    @staticmethod
    def get_bastion_username():
        return fablib.fablib_object.bastion_username

    @staticmethod
    def get_bastion_key_filename():
        return fablib.fablib_object.bastion_key_filename

    @staticmethod
    def get_bastion_public_addr():
        return fablib.fablib_object.bastion_public_addr

    @staticmethod
    def get_bastion_private_ipv4_addr():
        return fablib.fablib_object.bastion_private_ipv4_addr

    @staticmethod
    def get_bastion_private_ipv6_addr():
        return fablib.fablib_object.bastion_private_ipv6_addr

    @staticmethod
    def set_slice_manager(slice_manager):
        fablib.fablib_object.slice_manager = slice_manager

    @staticmethod
    def get_slice_manager():
        return fablib.fablib_object.slice_manager

    @staticmethod
    def create_slice_manager():
        return fablib.fablib_object.create_slice_manager()

    @staticmethod
    def new_slice(name):
        #fabric = fablib()
        from fabrictestbed_extensions.fablib.slice import Slice
        return Slice.new_slice(name=name)

    @staticmethod
    def get_site_advertisment(site):
        return_status, topology = fablib.get_slice_manager().resources()
        if return_status != Status.OK:
            raise Exception("Failed to get advertised_topology: {}, {}".format(return_status, topology))


        return topology.sites[site]

    @staticmethod
    def get_available_resources():
        return_status, topology = fablib.get_slice_manager().resources()
        if return_status != Status.OK:
            raise Exception("Failed to get advertised_topology: {}, {}".format(return_status, topology))

        return topology

    @staticmethod
    def get_slices(excludes=[SliceState.Dead,SliceState.Closing], verbose=False):
        from fabrictestbed_extensions.fablib.slice import Slice
        return_status, slices = fablib.get_slice_manager().slices(excludes=excludes)

        return_slices = []
        if return_status == Status.OK:
            for slice in slices:
                #print("{}:".format(slice.slice_name))
                #print("   ID         : {}".format(slice.slice_id))
                #print("   State      : {}".format(slice.slice_state))
                #print("   Lease End  : {}".format(slice.lease_end))
                #print()
                return_slices.append(Slice.get_slice(sm_slice=slice, load_config=False))
        else:
            print(f"Failure: {slices}")

        return return_slices

    @staticmethod
    def get_slice(name=None, slice_id=None, verbose=False):
        slices = fablib.get_slices()

        for slice in slices:
            if name != None and slice.get_name() == name:
                return slice
            if slice_id != None and slice.get_slice_id() == slice_id:
                return slice

        #Should not get here if the slice is found
        if name != None:
            raise Exception(f"Slice {name} not found")
        elif slice_id != None:
            raise Excption(f"Slice {slice_id} not found")
        else:
            raise Exception(f"get_slice missing slice_id or name")

    @staticmethod
    def delete_slice(slice_name=None):
        slice = fablib.get_slice(slice_name)
        slice.delete()

    @staticmethod
    def delete_all():
        slices = fablib.get_slices()

        for slice in slices:
            try:
                print(f"Deleting slice {slice.get_name()}", end='')
                slice.delete()
                print(f", Success!")
            except Exception as e:
                print(f", Failed!")



#init fablib object
fablib.fablib_object = fablib()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, CapacityHints, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

#from fabrictestbed_extensions.fabricx.fabricx import FabricX
#from fabrictestbed_extensions.fabricx.slicex import SliceX
#from fabrictestbed_extensions.fabricx.nodex import NodeX
#from .slicex import SliceX
#from .nodex import NodeX
#from .fabricx import FabricX


from ipaddress import ip_address, IPv4Address


#from fim.user import node


#from .abc_fablib import AbcFabLIB

from .. import images


#class Interface(AbcFabLIB):
class Interface():
    def __init__(self, component=None, fim_interface=None):
        """
        Constructor
        :return:
        """
        super().__init__()
        self.fim_interface  = fim_interface
        self.component = component

    def get_os_interface(self):
        try:
            os_iface = self.get_physical_os_interface()['ifname']
            vlan = self.get_vlan()

            if vlan != None:
                os_iface = f"{os_iface}.{vlan}"
        except:
            os_iface = None


        return os_iface

    def get_mac(self):
        try:
            os_iface = self.get_physical_os_interface()
            mac = os_iface['mac']
        except:
            mac = None

        return mac


    def get_physical_os_interface(self):


        if self.get_network() == None:
            return None

        network_name = self.get_network().get_name()
        node_name = self.get_node().get_name()

        #print(f"get_physical_os_interface: network_name = {network_name}")
        #print(f"get_physical_os_interface: node_name = {node_name}")


        return self.get_slice().get_interface_map()[network_name][node_name]

    def config_vlan_iface(self):
        if self.get_vlan() != None:
            self.get_node().add_vlan_os_interface(os_iface=self.get_physical_os_interface()['ifname'],
                                                  vlan=self.get_vlan())


    def set_ip(self, ip=None, cidr=None):
        self.get_node().set_ip_os_interface(os_iface=self.get_physical_os_interface()['ifname'],
                                            vlan=self.get_vlan(),
                                            ip=ip, cidr=cidr)

    def set_vlan(self, vlan=None):

        if_labels = self.get_fim_interface().get_property(pname="labels")
        if_labels.vlan = str(vlan)
        self.get_fim_interface().set_properties(labels=if_labels)

    def get_fim_interface(self):
        return self.fim_interface

    def get_bandwidth(self):
        return self.get_fim_interface().capacities.bw

    def get_vlan(self):
        try:
            vlan = self.get_fim_interface().get_property(pname="labels").vlan
        except:
            vlan = None
        return vlan

    def get_name(self):
        return self.get_fim_interface().name

    def get_component(self):
        return self.component

    def get_model(self):
        return self.get_component().get_model()

    def get_site(self):
        return self.get_component().get_site()

    def get_slice(self):
        return self.get_node().get_slice()

    def get_node(self):
        return self.get_component().get_node()

    def get_network(self):
        for net in self.get_slice().get_l2networks():
            if net.has_interface(self):
                return net

        return None
        #raise Exception(f"Network not found: interface {self.get_name()}")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import os
import traceback
import time

from abc import ABC, abstractmethod
from typing import List

from fabric_cf.orchestrator.orchestrator_proxy import SliceState
from fabrictestbed.slice_manager import SliceManager, Status, SliceState
from fabrictestbed.slice_editor import ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog

class AbcFabLIB(ABC):

    #credmgr_host = os.environ['FABRIC_CREDMGR_HOST']
    #orchestrator_host = os.environ['FABRIC_ORCHESTRATOR_HOST']
    #fabric_token=os.environ['FABRIC_TOKEN_LOCATION']

    #fabric_slice_private_key_passphrase = os.environ['FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE']

    #Basstion host setup
    #bastion_username = os.environ['FABRIC_BASTION_USERNAME']
    #bastion_key_filename = os.environ['FABRIC_BASTION_KEY_LOCATION']
    #bastion_public_addr = os.environ['FABRIC_BASTION_HOST']
    #bastion_private_ipv4_addr = os.environ['FABRIC_BASTION_HOST_PRIVATE_IPV4']
    #bastion_private_ipv6_addr = os.environ['FABRIC_BASTION_HOST_PRIVATE_IPV6']

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

        self.credmgr_host = None
        self.orchestrator_host = None
        self.fabric_token = None

        self.bastion_username = None
        self.bastion_key_filename = None
        self.bastion_public_addr = None
        self.bastion_private_ipv4_addr = '0.0.0.0'
        self.bastion_private_ipv6_addr = '0:0:0:0:0:0'

        #self.slice_public_key = None
        #self.slice_public_key_file = None
        #self.slice_private_key = None
        #self.slice_private_key_file = None
        #self.slice_private_key_passphrase = None

        self.slice_keys = {}
        self.default_slice_key = {}
        self.slice_keys['default'] = self.default_slice_key


        if "FABRIC_CREDMGR_HOST" in os.environ:
            self.credmgr_host = os.environ['FABRIC_CREDMGR_HOST']

        if "FABRIC_ORCHESTRATOR_HOST" in os.environ:
            self.orchestrator_host = os.environ['FABRIC_ORCHESTRATOR_HOST']

        if "FABRIC_TOKEN_LOCATION" in os.environ:
            self.fabric_token=os.environ['FABRIC_TOKEN_LOCATION']

        #Basstion host setup
        if "FABRIC_BASTION_USERNAME" in os.environ:
            self.bastion_username = os.environ['FABRIC_BASTION_USERNAME']
        if "FABRIC_BASTION_KEY_LOCATION" in os.environ:
            self.bastion_key_filename = os.environ['FABRIC_BASTION_KEY_LOCATION']
        if "FABRIC_BASTION_HOST" in os.environ:
            self.bastion_public_addr = os.environ['FABRIC_BASTION_HOST']
        if "FABRIC_BASTION_HOST_PRIVATE_IPV4" in os.environ:
            self.bastion_private_ipv4_addr = os.environ['FABRIC_BASTION_HOST_PRIVATE_IPV4']
        if "FABRIC_BASTION_HOST_PRIVATE_IPV6" in os.environ:
            self.bastion_private_ipv6_addr = os.environ['FABRIC_BASTION_HOST_PRIVATE_IPV6']

        #Slice Keys
        if "FABRIC_SLICE_PUBLIC_KEY_FILE" in os.environ:
            self.default_slice_key['slice_public_key_file'] = os.environ['FABRIC_SLICE_PUBLIC_KEY_FILE']
            with open(os.environ['FABRIC_SLICE_PUBLIC_KEY_FILE'], "r") as fd:
                self.default_slice_key['slice_public_key'] = fd.read().strip()
        if "FABRIC_SLICE_PRIVATE_KEY_FILE" in os.environ:
            #self.slice_private_key_file=os.environ['FABRIC_SLICE_PRIVATE_KEY_FILE']
            self.default_slice_key['slice_private_key_file'] = os.environ['FABRIC_SLICE_PRIVATE_KEY_FILE']
        if "FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE" in os.environ:
            #self.slice_private_key_passphrase = os.environ['FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE']
            self.default_slice_key['slice_private_key_passphrase'] = os.environ['slice_private_key_passphrase']



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import os
import traceback
import time

from abc import ABC, abstractmethod
from typing import List

from fabric_cf.orchestrator.orchestrator_proxy import SliceState
from fabrictestbed.slice_manager import SliceManager, Status, SliceState
from fabrictestbed.slice_editor import ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from ipaddress import ip_address, IPv4Address


class AbcTest(ABC):

    def __init__(self):
        self.slice_manager = SliceManager()
        self.advertised_topology = None
        self.slice = None
        self.slices = []
        self.topology = None

        self.bastion_username = os.environ['FABRIC_BASTION_USERNAME']
        self.bastion_keyfile = ''

        self.bastion_public_addr = 'bastion-1.fabric-testbed.net'
        self.bastion_private_ipv4_addr = '192.168.11.226'
        self.bastion_private_ipv6_addr = '2600:2701:5000:a902::c'

        #self.bastion_key_filename = '/Users/pruth/FABRIC/TESTING/pruth_fabric_rsa'
        self.bastion_key_filename = os.environ['HOME'] + "/.ssh/pruth_fabric_rsa"
        self.bastion_key_filename = os.environ['FABRIC_BASTION_KEY_LOCATION']

        self.node_ssh_key_priv_file=os.environ['SLICE_PRIVATE_KEY_FILE']
        self.node_ssh_key_pub_file=os.environ['SLICE_PUBLIC_KEY_FILE']

        self.node_ssh_key = None
        with open(self.node_ssh_key_pub_file, "r") as fd:
            self.node_ssh_key = fd.read().strip()

        self.pull_advertised_topology()

    def pull_advertised_topology(self):
        return_status, self.advertised_topology = self.slice_manager.resources()
        if return_status != Status.OK:
            print("Failed to get advertised_topology: {}".format(self.advertised_topology))


    def get_slice(self,slice_name=None, slice_id=None, slice_manager=None):
        if not slice_manager:
            slice_manager = AbcUtils.create_slice_manager()

        return_status, slices = slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])
        if return_status != Status.OK:
            raise Exception("Failed to get slices: {}".format(slices))
        try:

            if slice_id:
                slice = list(filter(lambda x: x.slice_id == slice_id, slices))[0]
            elif slice_name:
                slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
            else:
                raise Exception("Slice not found. Slice name or id requried. name: {}, slice_id: {}".format(str(slice_name),str(slice_id)))
        except:
            raise Exception("Slice not found name: {}, slice_id: {}".format(str(slice_name),str(slice_id)))

        return slice


    def wait_for_slice(self,slice,timeout=360,interval=10,progress=False):
        self.slice = slice
        timeout_start = time.time()

        if progress: print("Waiting for slice {} .".format(self.slice.slice_name), end = '')
        while time.time() < timeout_start + timeout:
            return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])

            if return_status == Status.OK:
                slice = list(filter(lambda x: x.slice_name == self.slice.slice_name, slices))[0]
                if slice.slice_state == "StableOK":
                    if progress: print(" Slice state: {}".format(slice.slice_state))
                    return slice
                if slice.slice_state == "Closing" or slice.slice_state == "Dead" or slice.slice_state == "StableError":
                    if progress: print(" Slice state: {}".format(slice.slice_state))
                    return slice
            else:
                print(f"Failure: {slices}")

            if progress: print(".", end = '')
            time.sleep(interval)

        if time.time() >= timeout_start + timeout:
            if progress: print(" Timeout exceeded ({} sec). Slice: {} ({})".format(timeout,slice.slice_name,slice.slice_state))
            return slice

    @staticmethod
    def validIPAddress(IP: str) -> str:
        try:
            return "IPv4" if type(ip_address(IP)) is IPv4Address else "IPv6"
        except ValueError:
            return "Invalid"

    def run_ssh_test(self, slice):

        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("run_ssh_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        for node_name, node in topology.nodes.items():
            #print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            #print("Node:")
            print(" Test: ssh | " + str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            try:
                #print("   Name              : {}".format(node_name))
                #print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
                #print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
                #print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
                #print("   Image             : {}".format(node.image_ref))
                #print("   Image Type        : {}".format(node.image_type))
                #print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
                #print("   Site              : {}".format(node.site))
                #print("   Management IP     : {}".format(node.management_ip))
                #print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
                #print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
                #print("   Components        : {}".format(node.components))
                #print("   Interfaces        : {}".format(node.interfaces))
                #print()

                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = 'Hello, FABRIC from node ' + hostname
                #print("expected_stdout: XX{}XX".format(expected_stdout.replace('\n','')))

                #print("------------------------- Test Output ---------------------------")
                script= '#!/bin/bash  \n' \
                'echo Hello, FABRIC from node `hostname -s`   \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n','')) + " | ", end = '')
                #print("-----------------------------------------------------------------")
                #stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    #print('Fail: --{}--  --{}--'.format(expected_stdout,stdout_str))
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()


    def open_ssh_client_direct(self, node_username, node, timeout=10,interval=10,progress=True):
        import paramiko
        timeout_start = time.time()

        client = None
        if progress: print("Waiting for ssh client connection {} .".format(node.name), end = '')
        while time.time() < timeout_start + timeout:

            try:
                management_ip = str(node.get_property(pname='management_ip'))
                #print("Node {0} IP {1}".format(node.name, management_ip))

                key = paramiko.RSAKey.from_private_key_file(self.node_ssh_key_priv_file)

                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=node_username,pkey = key)

                break
            except Exception as e:
                #print (str(e))
                #return str(e)
                pass

            if progress: print(".", end = '')
            time.sleep(interval)

        return client

    def open_ssh_client(self, node_username, node):
        import paramiko

        try:
            management_ip = str(node.get_property(pname='management_ip'))
            #print("Node {0} IP {1}".format(node.name, management_ip))

            key = paramiko.RSAKey.from_private_key_file(self.node_ssh_key_priv_file)

            bastion=paramiko.SSHClient()
            bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            bastion.connect(self.bastion_public_addr, username=self.bastion_username, key_filename=self.bastion_key_filename)

            bastion_transport = bastion.get_transport()
            if self.validIPAddress(management_ip) == 'IPv4':
                src_addr = (self.bastion_private_ipv4_addr, 22)
            elif self.validIPAddress(management_ip) == 'IPv6':
                src_addr = (self.bastion_private_ipv6_addr, 22)
            else:
                return 'Management IP Invalid: {}'.format(management_ip)

            dest_addr = (management_ip, 22)
            bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)


            client = paramiko.SSHClient()
            client.load_system_host_keys()
            client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            client.connect(management_ip,username=node_username,pkey = key, sock=bastion_channel)

        except Exception as e:
            return str(e)

        return client

    def close_ssh_client(self, ssh_client):
        import paramiko

        ssh_client.close()

    def execute_script(self, node_username, node, script):
        import paramiko

        try:
            management_ip = str(node.get_property(pname='management_ip'))
            #print("Node {0} IP {1}".format(node.name, management_ip))

            key = paramiko.RSAKey.from_private_key_file(self.node_ssh_key_priv_file)

            bastion=paramiko.SSHClient()
            bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            bastion.connect(self.bastion_public_addr, username=self.bastion_username, key_filename=self.bastion_key_filename)

            bastion_transport = bastion.get_transport()
            if self.validIPAddress(management_ip) == 'IPv4':
                src_addr = (self.bastion_private_ipv4_addr, 22)
            elif self.validIPAddress(management_ip) == 'IPv6':
                src_addr = (self.bastion_private_ipv6_addr, 22)
            else:
                return 'Management IP Invalid: {}'.format(management_ip)

            dest_addr = (management_ip, 22)
            bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)


            client = paramiko.SSHClient()
            client.load_system_host_keys()
            client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            client.connect(management_ip,username=node_username,pkey = key, sock=bastion_channel)

            stdin, stdout, stderr = client.exec_command('echo \"' + script + '\" > script.sh; chmod +x script.sh; sudo ./script.sh')
            stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
            #print ('')
            #print (str(stdout.read(),'utf-8').replace('\\n','\n'))
            #print (stdout_str)

            client.close()
        except Exception as e:
            return str(e)

        return stdout_str


    @abstractmethod
    def run(self, create_slice=True, run_test=True, delete=True):
        """
        Run the test
        :return:
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, CapacityHints, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

from ipaddress import ip_address, IPv4Address

#from .abc_fablib import AbcFabLIB

from .. import images


#class NetworkService(AbcFabLIB):
class NetworkService():
    network_service_map = { 'L2Bridge': ServiceType.L2Bridge,
                            'L2PTP': ServiceType.L2PTP,
                            'L2STS': ServiceType.L2STS,
                            }

    #Type names used in fim network services
    fim_network_service_types = [ 'L2Bridge', 'L2PTP', 'L2STS']


    @staticmethod
    def calculate_l2_nstype(interfaces=None):

        sites = set([])
        for interface in interfaces:
            sites.add(interface.get_site())

        rtn_nstype = None
        if len(sites) == 1:
            rtn_nstype = NetworkService.network_service_map['L2Bridge']
        elif len(sites) == 2 and len(interfaces) == 2:
            rtn_nstype = NetworkService.network_service_map['L2PTP']
        elif len(sites) == 2  and len(interfaces) > 2:
            rtn_nstype = NetworkService.network_service_map['L2STS']
        else:
            raise Exception(f"Invalid Network Service: Networks are limited to 2 unique sites. Site requested: {sites}")

        return rtn_nstype

    @staticmethod
    def validate_nstype(type, interfaces):

        sites = set([])
        nics = set([])
        for interface in interfaces:
            sites.add(interface.get_site())
            nics.add(interface.get_model())

        #models: 'NIC_Basic', 'NIC_ConnectX_6', 'NIC_ConnectX_5'
        if type == NetworkService.network_service_map['L2Bridge']:
            if not len(sites) == 1:
                raise Exception(f"Network type {type} must include interfaces from exactly one site. {len(sites)} sites requested: {sites}")

        elif type == NetworkService.network_service_map['L2PTP']:
            if not len(sites) == 2:
                raise Exception(f"Network type {type} must include interfaces from exactly two sites. {len(sites)} sites requested: {sites}")
            if 'NIC_Basic' in nics:
                raise Exception(f"Network type {type} does not support interfaces of type 'NIC_Basic'")

        elif type == NetworkService.network_service_map['L2STS']:
            if not len(sites) == 2:
                raise Exception(f"Network type {type} must include interfaces from exactly two sites. {len(sites)} sites requested: {sites}")
        else:
            raise Exception(f"Invalid l2 network type: {type}. Please choose from {NetworkService.fim_network_service_types} or None for automatic selection")

        return True


    @staticmethod
    def new_l2network(slice=None, name=None, interfaces=[], type=None):
        if type == None:
            nstype=NetworkService.calculate_l2_nstype(interfaces=interfaces)
        else:
            if type in NetworkService.fim_network_service_types:
                nstype=NetworkService.network_service_map[type]
            else:
                raise Exception(f"Invalid l2 network type: {type}. Please choose from {NetworkService.fim_network_service_types} or None for automatic selection")

        #validate nstype and interface List
        NetworkService.validate_nstype(nstype, interfaces)

        return NetworkService.new_network_service(slice=slice, name=name, nstype=nstype, interfaces=interfaces)

    @staticmethod
    def new_network_service(slice=None, name=None, nstype=None, interfaces=[]):
        fim_interfaces = []
        for interface in interfaces:
            fim_interfaces.append(interface.get_fim_interface())

        fim_network_service = slice.topology.add_network_service(name=name,
                                                                 nstype=nstype,
                                                                 interfaces=fim_interfaces)

        return NetworkService(slice = slice, fim_network_service = fim_network_service)

    @staticmethod
    def get_l2network_services(slice=None):
        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        for net_name, net in topology.network_services.items():
            if str(net.get_property('type')) in NetworkService.fim_network_service_types:
                rtn_network_services.append(NetworkService(slice = slice, fim_network_service = net))

        return rtn_network_services

    @staticmethod
    def get_l2network_service(slice=None, name=None):

        for net in NetworkService.get_l2network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.name}, network {name}")


    def __init__(self, slice=None, fim_network_service=None):
        """
        Constructor
        :return:
        """
        super().__init__()
        self.fim_network_service  = fim_network_service
        self.slice = slice

    def get_slice(self):
        return self.slice

    def get_fim_network_service(self):
        return self.fim_network_service

    def get_name(self):
        return self.get_fim_network_service().name

    def get_interfaces(self):
        interfaces = []
        for interface in self.get_fim_network_service().interface_list:
            interfaces.append(self.get_slice().get_interface(name=interface.name))

        return interfaces

    def get_interface(self, name=None):
        #print(f"network_service.get_interface: name {name}")
        for interface in self.get_interfaces():
            #print(f"network_service.get_interface: self.get_name() {self.get_name()}, interface.get_name(): {interface.get_name()}")

            #interface_name = f"{self.get_name()}-{interface.get_name()}"
            interface_name = f"{interface.get_name()}"
            #print(f"network_service.get_interface: interface_name {interface_name}, name: {name}")

            if interface_name == name:
                #print(f"returning iface: {interface.get_name()}")
                return interface

        return None
        #raise Exception(f"Interface not found: interface {name}")

    def has_interface(self, interface):

        if self.get_interface(name=interface.get_name()) == None:
            return False
        else:
            return True

    def find_nic_mapping(self, net_name, nodes):
        return_data = {}

        #copy scripts to nodes
        for node in nodes:
            #config node1
            file_attributes = upload_file(username, node, 'scripts/host_set_all_dataplane_ips.py','host_set_all_dataplane_ips.py')
            #print("file_attributes: {}".format(file_attributes))
            file_attributes = upload_file(username, node, 'scripts/find_nic_mapping.py','find_nic_mapping.py')
            #print("file_attributes: {}".format(file_attributes))


        #Config target node
        target_node = nodes[0]
        stdout = execute_script(username, target_node, 'sudo python3 host_set_all_dataplane_ips.py')
        #print("stdout: {}".format(stdout))
        #print(stdout)

        target_ifaces = json.loads(stdout.replace('\n',''))
        #print(ifaces)
        #for i in ifaces['management']:
        #    print(i)

        #for i in ifaces['dataplane']:
        #    print(i)
        #node1_map = { 'data' : {'ens6': '192.168.1.100', etc... }, 'management': { 'ens3': '10.1.1.1'}  }


        # Test s1 ifaces
        target_net_ip = None
        for node in nodes:
            #skip target node
            if node == target_node:
                continue

            #test node interfaces against target
            for target, ip in target_ifaces['dataplane']:
                node1_dataplane_ip = ip
                node2_dataplane_ip = ip.replace('100','101')
                node2_cidr = '24'
                #node2_management_ip = str(node2.management_ip)
                #print("S1 Name        : {}".format(node2.name))
                #print("Management IP    : {}".format(node2_management_ip))

                stdout = execute_script(username, node, 'python3 find_nic_mapping.py {} {} {} {}'.format('net',node2_dataplane_ip,node2_cidr,node1_dataplane_ip) )
                iface =  stdout.replace('\n','')
                #print("stdout: {}".format(stdout))
                if iface != 'None':
                    #print(iface)
                    #return iface
                    if target_net_ip == None:
                        return_data[target_node.name]=target
                    return_data[node.name]=iface
                    break



        return return_data

    def flush_dataplane_ips(nodes):

        for node in nodes:
            #config node1
            file_attributes = upload_file(username, node, 'scripts/host_flush_all_dataplane_ips.py','host_flush_all_dataplane_ips.py')
            #print("file_attributes: {}".format(file_attributes))

            stdout = execute_script(username, node, 'sudo python3 host_flush_all_dataplane_ips.py')
            #print("stdout: {}".format(stdout))
            #print(stdout)


    def flush_all_dataplane_ips():
        for node_name, node in experiment_topology.nodes.items():
            flush_dataplane_ips([node])



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState


from .abc_utils import AbcUtils

from .. import images


class NodeUtils(AbcUtils):

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()


    @staticmethod
    def validIPAddress(IP: str) -> str:
        try:
            return "IPv4" if type(ip_address(IP)) is IPv4Address else "IPv6"
        except ValueError:
            return "Invalid"

    @staticmethod
    def execute_script(node_username, node, script):
        import paramiko

        management_ip = str(node.get_property(pname='management_ip'))
        print("Node {0} IP {1}".format(node.name, management_ip))

        key = paramiko.RSAKey.from_private_key_file(AbcUtils.node_ssh_key_priv_file)

        bastion=paramiko.SSHClient()
        bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        bastion.connect(AbcUtils.bastion_public_addr, username=AbcUtils.bastion_username, key_filename=AbcUtils.bastion_key_filename)

        bastion_transport = bastion.get_transport()
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (self.bastion_private_ipv4_addr, 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (self.bastion_private_ipv6_addr, 22)
        else:
            print ('Management IP Invalid: {}'.format(management_ip))
            return

        dest_addr = (management_ip, 22)
        bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)


        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        client.connect(management_ip,username=node_username,pkey = key, sock=bastion_channel)

        stdin, stdout, stderr = client.exec_command('echo \"' + script + '\" > script.sh; chmod +x script.sh; sudo ./script.sh')
        print ('')
        print (str(stdout.read(),'utf-8').replace('\\n','\n'))

        client.close()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

import ipycytoscape as cy
from IPython.display import display
from ipywidgets import Output

from .abc_topology_editor import AbcTopologyEditor

from .. import images


class CytoscapeTopologyEditor(AbcTopologyEditor):
    # FABRIC design elements https://fabric-testbed.net/branding/style/
    FABRIC_PRIMARY = '#27aae1'
    FABRIC_PRIMARY_LIGHT = '#cde4ef'
    FABRIC_PRIMARY_DARK = '#078ac1'
    FABRIC_SECONDARY = '#f26522'
    FABRIC_SECONDARY_LIGHT = '#ff8542'
    FABRIC_SECONDARY_DARK = '#d24502'
    FABRIC_BLACK = '#231f20'
    FABRIC_DARK = '#433f40'
    FABRIC_GREY = '#666677'
    FABRIC_LIGHT = '#f3f3f9'
    FABRIC_WHITE = '#ffffff'
    FABRIC_LOGO = "fabric_logo.png"

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

        self.out = Output()

        self.cytoscapeobj = cy.CytoscapeWidget()
        self.data = { 'nodes': [], 'edges': [] }

        self.style = 'secondary'

    def toggle_style(self):

        if self.style == 'primary':
            self.style = 'secondary'
            color = self.FABRIC_SECONDARY
            dark_color = self.FABRIC_SECONDARY_DARK
        elif self.style == 'secondary':
            self.style = 'primary'
            color = self.FABRIC_PRIMARY
            dark_color = self.FABRIC_PRIMARY_DARK

        self.cytoscapeobj.set_style([{
                        'selector': 'node',
                        'css': {
                            'content': 'data(name)',
                            'text-valign': 'center',
                            'color': 'white',
                            'text-outline-width': 2,
                            'text-outline-color': dark_color,
                            'background-color': color
                        }
                        },
                        {
                        'selector': ':selected',
                        'css': {
                            'background-color': dark_color,
                            'line-color': dark_color,
                            'target-arrow-color': color,
                            'source-arrow-color': color,
                            'text-outline-color': color
                        }}
                        ])




    def build_data(self):
        cy_nodes = self.data['nodes']
        cy_edges = self.data['edges']

        # Build Site
        for site_name, site in self.advertised_topology.sites.items():
            print("site_name: {}".format(site_name))
            cy_nodes.append({ 'data': { 'id': site_name, 'name': site_name, 'href': 'http://cytoscape.org' }})

        #cy_edges.append({'data': { 'source': 'RENC', 'target': 'UKY' }})
        #cy_edges.append({'data': { 'source': 'UKY', 'target': 'LBNL' }})
        #cy_edges.append({'data': { 'source': 'LBNL', 'target': 'RENC' }})

        for link_name, link in self.advertised_topology.links.items():
            print("link_name {}, {}".format(link_name,link))
            print("\n\n Interfaces {}".format(link.interface_list))

            #Source
            source_interface = link.interface_list[0]
            source_parent = self.advertised_topology.get_parent_element(source_interface)
            source_node=self.advertised_topology.get_owner_node(source_parent)

            #Target
            target_interface = link.interface_list[1]
            target_parent = self.advertised_topology.get_parent_element(target_interface)
            target_node=self.advertised_topology.get_owner_node(target_parent)

            #Build edge
            cy_edges.append({'data': { 'source': source_node.name, 'target': target_node.name }})

    def setup_interaction(self):
        #out = Output()
        self.cytoscapeobj.on('node', 'click', self.on_click)
        self.cytoscapeobj.on('node', 'mouseover', self.on_mouseover)

    def on_click(self, node):
        with self.out:
            print('click: {}'.format(str(node)))
            self.toggle_style()

    def on_mouseover(self, node):
        with self.out:
            print('mouseovers: {}'.format(str(node)))


    def start(self):
        """
        Start the cytoeditor editors
        :return:
        """

        self.toggle_style()
        self.setup_interaction()
        self.build_data()
        self.cytoscapeobj.graph.add_graph_from_json(self.data)

        display(self.cytoscapeobj)
        display(self.out)

        #return self.out, self.cytoscapeobj



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, CapacityHints, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

#from fabrictestbed_extensions.fabricx.fabricx import FabricX
#from fabrictestbed_extensions.fabricx.slicex import SliceX
#from fabrictestbed_extensions.fabricx.nodex import NodeX
#from .slicex import SliceX
#from .nodex import NodeX
#from .fabricx import FabricX


from ipaddress import ip_address, IPv4Address


#from fim.user import node


#from .abc_fablib import AbcFabLIB

from .. import images


#class Component(AbcFabLIB):
class Component():
    component_model_map = { 'NIC_Basic': ComponentModelType.SharedNIC_ConnectX_6,
                            'NIC_ConnectX_6': ComponentModelType.SmartNIC_ConnectX_6,
                            'NIC_ConnectX_5': ComponentModelType.SmartNIC_ConnectX_5,
                            'NVME_P4510': ComponentModelType.NVME_P4510,
                            'GPU_TeslaT4': ComponentModelType.GPU_Tesla_T4,
                            'GPU_RTX6000': ComponentModelType.GPU_RTX6000
                            }


    @staticmethod
    def calculate_name(node=None, name=None):
        #Hack to make it possile to find interfaces
        return f"{node.get_name()}-{name}"

    @staticmethod
    def new_component(node=None, model=None, name=None):
        #Hack to make it possile to find interfaces
        name = Component.calculate_name(node=node, name=name)

        return Component(node = node, fim_component = node.fim_node.add_component(model_type=Component.component_model_map[model], name=name))
        #return Component(node = node, model=model, name=name)

    def __init__(self, node=None, fim_component=None):
        """
        Constructor
        :return:
        """
        super().__init__()
        self.fim_component = fim_component
        self.node = node

    def get_interfaces(self):
        from fabrictestbed_extensions.fablib.interface import Interface

        ifaces = []
        for fim_interface in self.get_fim_component().interface_list:
            ifaces.append(Interface(component=self, fim_interface=fim_interface))

        return ifaces

    def get_fim_component(self):
        return self.fim_component

    def get_slice(self):
        return self.node.get_slice()

    def get_node(self):
        return self.node

    def get_site(self):
        return self.node.get_site()

    def get_name(self):
        return self.get_fim_component().name

    def get_details(self):
        return self.get_fim_component().details

    def get_disk(self):
        return self.get_fim_component().get_property(pname='capacity_allocations').disk

    def get_unit(self):
        return self.get_fim_component().get_property(pname='capacity_allocations').unit

    def get_pci_addr(self):
        return self.get_fim_component().get_property(pname='label_allocations').bdf

    def get_model(self):
        #TODO: get new model names (NIC_Basic, etc.)
        return self.get_fim_model()

    def get_fim_model(self):
        return self.get_fim_component().model

    def get_type(self):
        return self.get_fim_component().type

    def configure_nvme(self, mount_point='/mnt/nvme_mount', verbose=False):
        output = []
        try:
            output.append(self.node.execute('sudo fdisk -l /dev/nvme*'))
            output.append(self.node.execute('sudo parted -s /dev/nvme0n1 mklabel gpt'))
            output.append(self.node.execute('sudo parted -s /dev/nvme0n1 print'))
            output.append(self.node.execute('sudo parted -s /dev/nvme0n1 print unit MB print free'))
            output.append(self.node.execute('sudo parted -s --align optimal /dev/nvme0n1 mkpart primary ext4 0% 960197MB'))
            output.append(self.node.execute('lsblk /dev/nvme0n1'))
            output.append(self.node.execute('sudo mkfs.ext4 /dev/nvme0n1p1'))
            output.append(self.node.execute(f'sudo mkdir {mount_point} && sudo mount /dev/nvme0n1p1 {mount_point}'))
            output.append(self.node.execute(f'df -h {mount_point}'))
        except Exception as e:
            print(f"config_nvme Fail: {self.get_name()}")
            traceback.print_exc()
            raise Exception(str(output))

        return output


class Disk(Component):

    def __init__(self, component):
        """
        Constructor
        :return:
        """
        super().__init__(component)

class NIC(Component):
    def __init__(self, component):
        """
        Constructor
        :return:
        """
        super().__init__(component)

class GPU(Component):
    def __init__(self, component):
        """
        Constructor
        :return:
        """
        super().__init__(component)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re
import json

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, CapacityHints, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

from ipaddress import ip_address, IPv4Address

#from .abc_fablib import AbcFabLIB
from fabrictestbed_extensions.fablib.fablib import fablib


from .. import images

#class Node(AbcFabLIB):
class Node():
    def __init__(self, slice, node):
        """
        Constructor
        :return:
        """
        super().__init__()
        self.fim_node = node
        self.slice = slice

        #Try to set the username.
        try:
            self.set_username()
        except:
            self.username = None

    @staticmethod
    def new_node(slice=None, name=None, site=None):
        from fabrictestbed_extensions.fablib.node import Node
        return Node(slice, slice.topology.add_node(name=name, site=site))

    @staticmethod
    def get_node(slice=None, node=None):
        from fabrictestbed_extensions.fablib.node import Node
        return Node(slice, node)

    def get_fim_node(self):
        return self.fim_node

    def set_capacities(self, cores=2, ram=2, disk=2):
        cap = Capacities(core=cores, ram=ram, disk=disk)
        self.get_fim_node().set_properties(capacities=cap)

    def set_instance_type(self, instance_type):
        self.get_fim_node().set_properties(capacity_hints=CapacityHints(instance_type=instance_type))

    def set_username(self, username=None):
        if 'centos' in self.get_image():
            self.username = 'centos'
        elif 'ubuntu' in self.get_image():
            self.username = 'ubuntu'
        else:
            self.username = None

    def set_image(self, image, username=None, image_type='qcow2'):
        self.get_fim_node().set_properties(image_type=image_type, image_ref=image)
        self.set_username(username=username)

    def set_host(self, host_name=None):
        #excample: host_name='renc-w2.fabric-testbed.net'
        labels = Labels()
        labels.instance_parent = host_name
        self.get_fim_node().set_properties(labels=labels)


    def get_slice(self):
        return self.slice

    def get_name(self):
        return self.get_fim_node().name

    def get_cores(self):
        return self.get_fim_node().get_property(pname='capacity_allocations').core

    def get_ram(self):
        return self.get_fim_node().get_property(pname='capacity_allocations').ram

    def get_disk(self):
        return self.get_fim_node().get_property(pname='capacity_allocations').disk

    def get_image(self):
        return self.get_fim_node().image_ref

    def get_image_type(self):
        return self.get_fim_node().image_type

    def get_host(self):
        return self.get_fim_node().get_property(pname='label_allocations').instance_parent

    def get_site(self):
        return self.get_fim_node().site

    def get_management_ip(self):
        return self.get_fim_node().management_ip

    def get_reservation_id(self):
        return self.get_fim_node().get_property(pname='reservation_info').reservation_id

    def get_reservation_state(self):
        return self.get_fim_node().get_property(pname='reservation_info').reservation_state

    def get_interfaces(self):
        from fabrictestbed_extensions.fablib.interface import Interface

        interfaces = []
        for component in self.get_components():
            for interface in component.get_interfaces():
                interfaces.append(interface)

        return interfaces

    def get_interface(self, name=None, network_name=None):
        from fabrictestbed_extensions.fablib.interface import Interface

        if name != None:
            for component in self.get_components():
                for interface in component.get_interfaces():
                    if interface.get_name() == name:
                        return interface
        elif network_name != None:
            for interface in self.get_interfaces():
                if interface != None and interface.get_network().get_name() == network_name:
                    return interface

        raise Exception("Interface not found: {}".format(name))


    def get_username(self):
        return self.username

    def get_public_key(self):
        return self.get_slice().get_slice_public_key()

    def get_public_key_file(self):
        return self.get_slice().get_slice_public_key_file()

    def get_private_key(self):
        return self.get_slice().get_slice_private_key()

    def get_private_key_file(self):
        return self.get_slice().get_slice_private_key_file()

    def get_private_key_passphrase(self):
        return self.get_slice().get_private_key_passphrase()

    def add_component(self, model=None, name=None):
        from fabrictestbed_extensions.fablib.component import Component
        return Component.new_component(node=self, model=model, name=name)

    def get_components(self):
        from fabrictestbed_extensions.fablib.component import Component

        return_components = []
        for component_name, component in self.get_fim_node().components.items():
            #return_components.append(Component(self,component))
            return_components.append(Component(self,component))

        return return_components

    def get_component(self, name, verbose=False):
        from fabrictestbed_extensions.fablib.component import Component
        try:
            name = Component.calculate_name(node=self, name=name)
            return Component(self,self.get_fim_node().components[name])
        except Exception as e:
            if verbose:
                traceback.print_exc()
            raise Exception(f"Component not found: {name}")


    def get_ssh_command(self):
        return 'ssh -i {} -J {}@{} {}@{}'.format(self.get_private_key_file(),
                                           fablib.get_bastion_username(),
                                           fablib.get_bastion_public_addr(),
                                           self.get_username(),
                                           self.get_management_ip())

    def validIPAddress(self, IP: str) -> str:
        try:
            return "IPv4" if type(ip_address(IP)) is IPv4Address else "IPv6"
        except ValueError:
            return "Invalid"

    def execute(self, command, retry=3, retry_interval=10):
        import paramiko
        import time

        #Get and test src and management_ips
        management_ip = str(self.get_fim_node().get_property(pname='management_ip'))
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (fablib.get_bastion_private_ipv4_addr(), 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (fablib.get_bastion_private_ipv6_addr(), 22)
        else:
            raise Exception(f"upload_file: Management IP Invalid: {management_ip}")
        dest_addr = (management_ip, 22)

        for attempt in range(retry):
            try:
                if self.get_private_key_passphrase():
                    key = paramiko.RSAKey.from_private_key_file(self.get_private_key_file(),  password=self.get_private_key_passphrase())
                else:
                    key = paramiko.RSAKey.from_private_key_file(self.get_private_key_file())

                bastion=paramiko.SSHClient()
                bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                bastion.connect(fablib.get_bastion_public_addr(), username=fablib.get_bastion_username(), key_filename=fablib.get_bastion_key_filename())

                bastion_transport = bastion.get_transport()
                bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)

                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=self.username,pkey = key, sock=bastion_channel)

                stdin, stdout, stderr = client.exec_command('echo \"' + command + '\" > script.sh; chmod +x script.sh; sudo ./script.sh')
                rtn_stdout = str(stdout.read(),'utf-8').replace('\\n','\n')
                rtn_stderr = str(stderr.read(),'utf-8').replace('\\n','\n')


                client.close()

                return rtn_stdout, rtn_stderr
                #success, skip other tries
                break
            except Exception as e:
                if attempt+1 == retry:
                    raise e

                #Fail, try again
                print(f"SSH execute fail. Slice: {self.get_slice().get_name()}, Node: {self.get_name()}, trying again")
                print(f"Fail: {e}")
                #traceback.print_exc()
                time.sleep(retry_interval)
                pass

        raise Exception("scp download failed")


    def upload_file(self, local_file_path, remote_file_path, retry=3, retry_interval=10):
        import paramiko
        import time

        #Get and test src and management_ips
        management_ip = str(self.get_fim_node().get_property(pname='management_ip'))
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (fablib.get_bastion_private_ipv4_addr(), 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (fablib.get_bastion_private_ipv6_addr(), 22)
        else:
            raise Exception(f"upload_file: Management IP Invalid: {management_ip}")
        dest_addr = (management_ip, 22)

        for attempt in range(retry):
            try:

                if self.get_private_key_passphrase():
                    key = paramiko.RSAKey.from_private_key_file(self.get_private_key_file(),  password=self.get_private_key_passphrase())
                else:
                    key = paramiko.RSAKey.from_private_key_file(self.get_private_key_file())

                bastion=paramiko.SSHClient()
                bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                bastion.connect(fablib.get_bastion_public_addr(), username=fablib.get_bastion_username(), key_filename=fablib.get_bastion_key_filename())

                bastion_transport = bastion.get_transport()
                bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)


                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=self.username,pkey = key, sock=bastion_channel)

                ftp_client=client.open_sftp()
                file_attributes = ftp_client.put(local_file_path, remote_file_path)
                ftp_client.close()

                return file_attributes
                #success, skip other tries
                break
            except Exception as e:
                if attempt+1 == retry:
                    raise e

                #Fail, try again
                print(f"SCP upload fail. Slice: {self.get_slice().get_name()}, Node: {self.get_name()}, trying again")
                print(f"Fail: {e}")
                #traceback.print_exc()
                time.sleep(retry_interval)
                pass

        raise Exception("scp upload failed")


    def download_file(self, local_file_path, remote_file_path, retry=3, retry_interval=10):
        import paramiko
        import time

        #Get and test src and management_ips
        management_ip = str(self.get_fim_node().get_property(pname='management_ip'))
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (fablib.get_bastion_private_ipv4_addr(), 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (fablib.get_bastion_private_ipv6_addr(), 22)
        else:
            raise Exception(f"upload_file: Management IP Invalid: {management_ip}")
        dest_addr = (management_ip, 22)

        for attempt in range(retry):
            try:
                if self.get_private_key_passphrase():
                    key = paramiko.RSAKey.from_private_key_file(self.get_private_key_file(),  password=self.get_private_key_passphrase())
                else:
                    key = paramiko.RSAKey.from_private_key_file(self.get_private_key_file())

                bastion=paramiko.SSHClient()
                bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                bastion.connect(fablib.get_bastion_public_addr(), username=fablib.get_bastion_username(), key_filename=fablib.get_bastion_key_filename())

                bastion_transport = bastion.get_transport()
                bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)

                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=self.username,pkey = key, sock=bastion_channel)


                ftp_client=client.open_sftp()
                file_attributes = ftp_client.get(local_file_path, remote_file_path)
                ftp_client.close()

                return file_attributes
                #success, skip other tries
                break
            except Exception as e:
                if attempt+1 == retry:
                    raise e

                #Fail, try again
                print(f"SCP download fail. Slice: {self.get_slice().get_name()}, Node: {self.get_name()}, trying again")
                print(f"Fail: {e}")
                #traceback.print_exc()
                time.sleep(retry_interval)
                pass

        raise Exception("scp download failed")


    def test_ssh(self):
        try:
            self.execute('ls', retry=1, retry_interval=10)
        except:
            return False
        return True

    def wait_for_ssh(self, retry=6, retry_interval=10):
        try:
            self.execute('echo hello, fabric', retry=retry, retry_interval=retry_interval)
        except:
            return False
        return True

    def get_management_os_interface(self):
        #Assumes that the default route uses the management network

        stdout, stderr = self.execute("sudo ip -j route list")
        stdout_json = json.loads(stdout)

        #print(pythonObj)
        for i in stdout_json:
            if i['dst'] == 'default':
                return  i['dev']

    def get_dataplane_os_interfaces(self):
        management_dev = self.get_management_os_interface()

        stdout, stderr = self.execute("sudo ip -j addr list")
        stdout_json = json.loads(stdout)
        dataplane_devs = []
        for i in stdout_json:
            if i['ifname'] != 'lo' and i['ifname'] !=  management_dev:
                dataplane_devs.append({'ifname': i['ifname'], 'mac': i['address']})

        return dataplane_devs

    def flush_all_os_interfaces(self):
        for iface in self.get_dataplane_os_interfaces():
            self.flush_os_interface(iface['ifname'])

    def flush_os_interface(self, os_iface):
        stdout, stderr = self.execute(f"sudo ip addr flush dev {os_iface}")

    def set_ip_os_interface(self, os_iface=None, vlan=None, ip=None, cidr=None, mtu=None):
        #Bring up base iface
        #print(f"node.set_ip_os_interface: os_iface {os_iface}, vlan {vlan}")
        command = f'sudo ip link set dev {os_iface} up'
        if mtu != None:
            command += f" mtu {mtu}"
        stdout, stderr = self.execute(command)

        #config vlan iface
        if vlan != None:
            #create vlan iface
            command = f'sudo ip link add link {os_iface} name {os_iface}.{vlan} type vlan id {vlan}'
            stdout, stderr = self.execute(command)

            #bring up vlan iface
            os_iface = f"{os_iface}.{vlan}"
            command = f'sudo ip link set dev {os_iface} up'
            if mtu != None:
                command += f" mtu {mtu}"
            stdout, stderr = self.execute(command)

        if ip != None and cidr != None:
            #Set ip
            command = f"sudo ip addr add {ip}/{cidr} dev {os_iface}"
            stdout, stderr = self.execute(command)

    def clear_all_ifaces(self):
        self.remove_all_vlan_os_interfaces()
        self.flush_all_os_interfaces()


    def remove_all_vlan_os_interfaces(self):
        management_os_iface = self.get_management_os_interface()

        stdout, stderr = self.execute("sudo ip -j addr list")
        stdout_json = json.loads(stdout)
        dataplane_devs = []
        for i in stdout_json:
            if i['ifname'] == management_os_iface or i['ifname'] == 'lo':
                stdout_json.remove(i)
                continue

            #If iface is vlan linked to base iface
            if 'link' in i.keys():
                self.remove_vlan_os_interface(os_iface=i['ifname'])

    def save_data(self):
        data = {}
        #Get interface data
        interfaces = {}
        for i in self.get_interfaces():
            #print(f"interface: {i.get_name()}")
            #print(f"os_interface: {i.get_physical_os_interface()}")
            if i.get_network() != None:
                network_name = i.get_network().get_name()
                #print(f"network: {i.get_network().get_name()}")
            else:
                network_name = None
                #print(f"network: None")

            interfaces[i.get_name()] =  { 'network':  network_name,
                         'os_interface':  i.get_physical_os_interface() }

        with open(f'{self.get_name()}.json', 'w') as outfile:
            json.dump(interfaces, outfile)

        #print(f"interfaces: {json.dumps(interfaces).replace('\"','\\"')}")

        self.upload_file(f'{self.get_name()}.json', f'{self.get_name()}.json')

    def load_data(self):

        self.download_file(f'{self.get_name()}.json', f'{self.get_name()}.json')

        interfaces=""
        with open(f'{self.get_name()}.json', 'r') as infile:
            interfaces = json.load(infile)


        interface_map = self.get_slice().network_iface_map #= self.get_slice().get_interface_map()
        #print(f"interfaces {interfaces}")
        for interface_name, net_map in interfaces.items():
            #print(f"interface_name: {net_map}:")
            if net_map['network'] != None:
                interface_map[net_map['network']][self.get_name()] = net_map['os_interface']

        self.get_slice().network_iface_map = interface_map

    def remove_vlan_os_interface(self, os_iface=None):
        command = f"sudo ip -j addr show {os_iface}"
        stdout, stderr = self.execute(command)
        try:
            [stdout_json] = json.loads(stdout)
        except Exception as e:
            print(f"os_iface: {os_iface}, stdout: {stdout}, stderr: {stderr}")
            raise e


        link = stdout_json['link']

        command = f"sudo ip link del link {link} name {os_iface}"
        stdout, stderr = self.execute(command)

    def add_vlan_os_interface(self, os_iface=None, vlan=None, ip=None, cidr=None, mtu=None):
        command = f'sudo ip link add link {os_iface} name {os_iface}.{vlan} type vlan id {vlan}'
        stdout, stderr = self.execute(command)
        command = f'sudo ip link set dev {os_iface}.{vlan} up'
        stdout, stderr = self.execute(command)

        if ip != None and cidr != None:
            self.set_ip_os_interface(os_iface=f"{os_iface}.{vlan}", ip=ip, cidr=cidr, mtu=mtu)

    def ping_test(self, dst_ip):
        command = f'ping -c 3 {dst_ip}  2>&1 > /dev/null && echo Success'
        stdout, stderr = self.execute(command)
        if stdout.replace("\n","") == 'Success':
            return True
        else:
            return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, CapacityHints, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

#from .slicex import SliceX
#from .nodex import NodeX
#from .fabricx import FabricX

#from .abc_fablib import AbcFabLIB

from .. import images

from fabrictestbed_extensions.fablib.fablib import fablib

class Slice():

    def __init__(self, name=None):
        """
        Constructor
        :return:
        """
        super().__init__()
        #print(f"Creating Slice: Name: {name}, Slice: {slice}")
        self.slice_name = name
        self.sm_slice = None
        self.slice_id = None
        self.topology = None

        self.slice_key = fablib.get_default_slice_key()

    @staticmethod
    def new_slice(name=None):
        """
        Create a new slice
        @param name slice name
        @return fablib Slice object
        """

        slice = Slice(name=name)
        slice.topology = ExperimentTopology()
        return slice

    @staticmethod
    def get_slice(sm_slice=None, verbose=False, load_config=True):
        """
        Create a new slice
        @param name slice name
        @return fablib Slice object
        """
        slice = Slice(name=sm_slice.slice_name)
        slice.sm_slice = sm_slice
        slice.slice_id = sm_slice.slice_id
        slice.slice_name = sm_slice.slice_name
        slice.topology = fablib.get_slice_manager().get_slice_topology(slice_object=slice.sm_slice)

        try:
            slice.update()
        except:
            if verbose:
                print(f"Slice {slice.slice_name} could not be updated: slice.get_slice")

        if load_config:
            try:
                slice.load_config()
            except:
                print(f"Slice {slice.slice_name} config could not loaded: slice.get_slice")

        return slice

    def get_fim_topology(self):
        return self.topology

    def update_slice(self):
        #Update slice
        return_status, slices = fablib.get_slice_manager().slices(excludes=[SliceState.Dead,SliceState.Closing])
        if return_status == Status.OK:
            self.sm_slice = list(filter(lambda x: x.slice_name == self.slice_name, slices))[0]
            self.slice_id = self.sm_slice.slice_id
        else:
            raise Exception("Failed to get slice list: {}, {}".format(return_status, slices))


    def update_topology(self):
        #Update topology
        return_status, new_topo = fablib.get_slice_manager().get_slice_topology(slice_object=self.sm_slice)
        if return_status != Status.OK:
            raise Exception("Failed to get slice topology: {}, {}".format(return_status, new_topo))

        #Set slice attibutes
        self.topology = new_topo

    def update(self):
        self.update_slice()
        self.update_topology()

    def get_slice_public_key(self):
        return self.slice_key['slice_public_key']

    def get_private_key_passphrase(self):
        if 'slice_private_key_passphrase' in self.slice_key.keys():
            return self.slice_key['slice_private_key_passphrase']
        else:
            return None

    def get_slice_public_key(self):
        if 'slice_public_key' in self.slice_key.keys():
            return self.slice_key['slice_public_key']
        else:
            return None

    def get_slice_public_key_file(self):
        if 'slice_public_key_file' in self.slice_key.keys():
            return self.slice_key['slice_public_key_file']
        else:
            return None

    def get_slice_private_key_file(self):
        if 'slice_private_key_file' in self.slice_key.keys():
            return self.slice_key['slice_private_key_file']
        else:
            return None

    def get_state(self):
        return self.sm_slice.slice_state

    def get_name(self):
        return self.slice_name

    def get_slice_id(self):
        return self.slice_id

    def get_lease_end(self):
        return self.sm_slice.lease_end

    def add_l2network(self, name=None, interfaces=[], type=None):
        from fabrictestbed_extensions.fablib.network_service import NetworkService
        return NetworkService.new_l2network(slice=self, name=name, interfaces=interfaces, type=type)

    def add_node(self, name, site):
        from fabrictestbed_extensions.fablib.node import Node
        return Node.new_node(slice=self, name=name, site=site)

    def get_nodes(self):
        from fabrictestbed_extensions.fablib.node import Node
        #self.update()

        return_nodes = []

        #fails for topology that does not have nodes
        try:
            for node_name, node in self.get_fim_topology().nodes.items():
                return_nodes.append(Node.get_node(self,node))
        except Exception as e:
            print("get_nodes: exception")
            traceback.print_exc()
            pass
        return return_nodes

    def get_node(self, name, verbose=False):
        from fabrictestbed_extensions.fablib.node import Node
        #self.update()
        try:
            return Node.get_node(self,self.get_fim_topology().nodes[name])
        except Exception as e:
            if verbose:
                traceback.print_exc()
            raise Exception(f"Node not found: {name}")

    def get_interfaces(self):
        interfaces = []
        for node in self.get_nodes():
            for interface in node.get_interfaces():
                interfaces.append(interface)
        return interfaces

    def get_interface(self, name=None):
        for interface in self.get_interfaces():
            if name.endswith(interface.get_name()):
                return interface

        raise Exception("Interface not found: {}".format(name))



    def get_l2networks(self, verbose=False):
        from fabrictestbed_extensions.fablib.network_service import NetworkService

        try:
            return NetworkService.get_l2network_services(self)
        except Exception as e:
            if verbose:
                traceback.print_exc()
        return None

    def get_l2network(self, name=None, verbose=False):
        from fabrictestbed_extensions.fablib.network_service import NetworkService

        try:
            return NetworkService.get_l2network_service(self,name)
        except Exception as e:
            if verbose:
                traceback.print_exc()
        return None


    def delete(self):
        return_status, result = fablib.get_slice_manager().delete(slice_object=self.sm_slice)

        if return_status != Status.OK:
            raise Exception("Failed to delete slice: {}, {}".format(return_status, result))

        self.topology = None

    def renew(self, end_date):

        return_status, result = fablib.get_slice_manager().renew(slice_object=self.sm_slice,
                                     new_lease_end_time = end_date)

        if return_status != Status.OK:
            raise Exception("Failed to renew slice: {}, {}".format(return_status, result))

    def wait(self, timeout=360,interval=10,progress=False):
        slice_name=self.sm_slice.slice_name
        slice_id=self.sm_slice.slice_id

        timeout_start = time.time()
        slice = self.sm_slice

        if progress: print("Waiting for slice .", end = '')
        while time.time() < timeout_start + timeout:
            return_status, slices = fablib.get_slice_manager().slices(excludes=[SliceState.Dead,SliceState.Closing])

            if return_status == Status.OK:
                slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
                if slice.slice_state == "StableOK":
                    if progress: print(" Slice state: {}".format(slice.slice_state))
                    return slice
                if slice.slice_state == "Closing" or slice.slice_state == "Dead" or slice.slice_state == "StableError":
                    if progress: print(" Slice state: {}".format(slice.slice_state))
                    return slice
            else:
                print(f"Failure: {slices}")

            if progress: print(".", end = '')
            time.sleep(interval)

        if time.time() >= timeout_start + timeout:
            if progress: print(" Timeout exceeded ({} sec). Slice: {} ({})".format(timeout,slice.slice_name,slice.slice_state))
            return slice

        #Update the fim topology (wait to avoid get topology bug)
        time.sleep(interval)
        self.update()

    def get_interface_map(self):
        if not hasattr(self, 'network_iface_map'):
            self.load_interface_map()

        return self.network_iface_map

    def post_boot_config(self, verbose=False):

        # Find the interface to network map
        self.build_interface_map()

        # Interface map in nodes
        for node in self.get_nodes():
            node.save_data()

        for interface in self.get_interfaces():
            interface.config_vlan_iface()

    def load_config(self):
        self.load_interface_map()

    def load_interface_map(self, verbose=False):
        self.network_iface_map = {}
        for net in self.get_l2networks():
            self.network_iface_map[net.get_name()] = {}

        for node in self.get_nodes():
            node.load_data()


    def build_interface_map(self, verbose=False):
        self.network_iface_map = {}
        for net in self.get_l2networks():
            iface_map = {}

            if verbose == True:
                print(f"Buiding iface map for network: {net.get_name()}")
            ifaces = net.get_interfaces()

            #target iface/node
            target_iface =  ifaces.pop()
            target_node = target_iface.get_node()
            target_os_ifaces = target_node.get_dataplane_os_interfaces()
            target_node.clear_all_ifaces()

            #print(f"{target_node.get_ssh_command()}")

            target_iface_nums = []
            for target_os_iface in target_os_ifaces:
                target_os_iface_name = target_os_iface['ifname']
                iface_num=target_os_ifaces.index(target_os_iface)+1
                target_node.set_ip_os_interface(os_iface=target_os_iface_name,
                                                  vlan=target_iface.get_vlan(),
                                                  ip=f'192.168.{iface_num}.1',
                                                  cidr = '24'
                                                 )
                target_iface_nums.append(iface_num)


            #print(f"target_iface: {target_iface.get_name()}")
            #print(f"target_iface.get_vlan(): {target_iface.get_vlan()}")
            #print(f"target_node: {target_node.get_name()}")
            #print(f"target_os_ifaces: {target_os_ifaces}")


            for iface in ifaces:
                node = iface.get_node()
                node.clear_all_ifaces()
                node_os_ifaces = node.get_dataplane_os_interfaces()


                #print(f"test_node: {node.get_name()}")
                #print(f"test_iface: {iface.get_name()}")
                #print(f"node_os_ifaces: {node_os_ifaces}")
                #print(f"iface.get_vlan(): {iface.get_vlan()}")
                #print(f"{node.get_ssh_command()}")

                found = False
                for node_os_iface in node_os_ifaces:
                    node_os_iface_name = node_os_iface['ifname']
                    #print(f"target_iface_nums: {target_iface_nums}")
                    for net_num in target_iface_nums:
                        dst_ip=f'192.168.{net_num}.1'

                        ip=f'192.168.{net_num}.2'

                        #set interface
                        node.set_ip_os_interface(os_iface=node_os_iface_name,
                                                 vlan=iface.get_vlan(),
                                                 ip=ip,
                                                 cidr='24')

                        #ping test
                        #print(f"ping test {node.get_name()}:{node_os_iface_name} ->  - {ip} to {dst_ip}")
                        test_result = node.ping_test(dst_ip)
                        #print(f"Ping test result: {test_result}")

                        if iface.get_vlan() == None:
                            node.flush_os_interface(node_os_iface_name)
                        else:
                            node.remove_vlan_os_interface(os_iface=f"{node_os_iface_name}.{iface.get_vlan()}")

                        if test_result:
                            #print(f"test_result true: {test_result}")
                            target_iface_nums = [ net_num ]
                            found = True
                            iface_map[node.get_name()] = node_os_iface
                            iface_map[target_node.get_name()] = target_os_ifaces[net_num-1]
                            break

                    if found:
                        break

            self.network_iface_map[net.get_name()] = iface_map
            target_node.clear_all_ifaces()

        if verbose:
            print(f"network_iface_map: {self.network_iface_map}")

    def submit(self, wait=False, wait_timeout=360, wait_interval=10, wait_progress=False, wait_ssh=False):
        from fabrictestbed_extensions.fablib.fablib import fablib
        fabric = fablib()

        # Generate Slice Graph
        slice_graph = self.get_fim_topology().serialize()

        # Request slice from Orchestrator
        return_status, slice_reservations = fablib.get_slice_manager().create(slice_name=self.slice_name,
                                                                slice_graph=slice_graph,
                                                                ssh_key=self.get_slice_public_key())
        if return_status != Status.OK:
            raise Exception("Failed to submit slice: {}, {}".format(return_status, slice_reservations))

        time.sleep(10)
        self.update_slice()

        if wait or wait_progress:
            self.wait(timeout=wait_timeout,interval=wait_interval,progress=wait_progress)

            if wait_progress:
                print("Running post boot config ...",end="")

            time.sleep(30)
            self.update()

            for node in self.get_nodes():
                node.wait_for_ssh()

            self.post_boot_config()

        if wait_progress:
            print("Done!")
